@@ -1,5 +1,6 @@
 # coding=utf-8
 """Shared utility functions"""
+
 import argparse
 import collections
 import functools
@@ -12,6 +13,9 @@ import subprocess
 import sys
 import threading
 import unicodedata
+from difflib import (
+    SequenceMatcher,
+)
 from enum import (
     Enum,
 )
@@ -41,12 +45,10 @@ from .argparse_custom import (
 if TYPE_CHECKING:  # pragma: no cover
     import cmd2  # noqa: F401
 
-    PopenTextIO = subprocess.Popen[bytes]
-
+    PopenTextIO = subprocess.Popen[str]
 else:
     PopenTextIO = subprocess.Popen
 
-
 _T = TypeVar('_T')
 
 
@@ -91,11 +93,14 @@ def strip_quotes(arg: str) -> str:
     return arg
 
 
-def str_to_bool(val: str) -> bool:
-    """Converts a string to a boolean based on its value.
+def to_bool(val: Any) -> bool:
+    """Converts anything to a boolean based on its value.
 
-    :param val: string being converted
-    :return: boolean value expressed in the string
+    Strings like "True", "true", "False", and "false" return True, True, False, and False
+    respectively. All other values are converted using bool()
+
+    :param val: value being converted
+    :return: boolean value expressed in the passed in value
     :raises: ValueError if the string does not contain a value corresponding to a boolean value
     """
     if isinstance(val, str):
@@ -103,7 +108,11 @@ def str_to_bool(val: str) -> bool:
             return True
         elif val.capitalize() == str(False):
             return False
-    raise ValueError("must be True or False (case-insensitive)")
+        raise ValueError("must be True or False (case-insensitive)")
+    elif isinstance(val, bool):
+        return val
+    else:
+        return bool(val)
 
 
 class Settable:
@@ -128,7 +137,7 @@ class Settable:
         :param name: name of the instance attribute being made settable
         :param val_type: callable used to cast the string value from the command line into its proper type and
                          even validate its value. Setting this to bool provides tab completion for true/false and
-                         validation using str_to_bool(). The val_type function should raise an exception if it fails.
+                         validation using to_bool(). The val_type function should raise an exception if it fails.
                          This exception will be caught and printed by Cmd.do_set().
         :param description: string describing this setting
         :param settable_object: object to which the instance attribute belongs (e.g. self)
@@ -149,13 +158,13 @@ class Settable:
         :param choices_provider: function that provides choices for this argument
         :param completer: tab completion function that provides choices for this argument
         """
-        if val_type == bool:
+        if val_type is bool:
 
             def get_bool_choices(_) -> List[str]:  # type: ignore[no-untyped-def]
                 """Used to tab complete lowercase boolean values"""
                 return ['true', 'false']
 
-            val_type = str_to_bool
+            val_type = to_bool
             choices_provider = cast(ChoicesProviderFunc, get_bool_choices)
 
         self.name = name
@@ -663,19 +672,21 @@ class ProcReader:
 
         # Run until process completes
         while self._proc.poll() is None:
-            # noinspection PyUnresolvedReferences
             available = read_stream.peek()  # type: ignore[attr-defined]
             if available:
                 read_stream.read(len(available))
                 self._write_bytes(write_stream, available)
 
     @staticmethod
-    def _write_bytes(stream: Union[StdSim, TextIO], to_write: bytes) -> None:
+    def _write_bytes(stream: Union[StdSim, TextIO], to_write: Union[bytes, str]) -> None:
         """
         Write bytes to a stream
         :param stream: the stream being written to
         :param to_write: the bytes being written
         """
+        if isinstance(to_write, str):
+            to_write = to_write.encode()
+
         try:
             stream.buffer.write(to_write)
         except BrokenPipeError:
@@ -863,7 +874,8 @@ def align_text(
     )
 
     if width is None:
-        width = shutil.get_terminal_size().columns
+        # Prior to Python 3.11 this can return 0, so use a fallback if needed.
+        width = shutil.get_terminal_size().columns or constants.DEFAULT_TERMINAL_WIDTH
 
     if width < 1:
         raise ValueError("width must be at least 1")
@@ -1179,9 +1191,7 @@ def get_defining_class(meth: Callable[..., Any]) -> Optional[Type[Any]]:
     if isinstance(meth, functools.partial):
         return get_defining_class(meth.func)
     if inspect.ismethod(meth) or (
-        inspect.isbuiltin(meth)
-        and getattr(meth, '__self__') is not None
-        and getattr(meth.__self__, '__class__')  # type: ignore[attr-defined]
+        inspect.isbuiltin(meth) and getattr(meth, '__self__') is not None and getattr(meth.__self__, '__class__')
     ):
         for cls in inspect.getmro(meth.__self__.__class__):  # type: ignore[attr-defined]
             if meth.__name__ in cls.__dict__:
@@ -1254,3 +1264,36 @@ def strip_doc_annotations(doc: str) -> str:
         elif found_first:
             break
     return cmd_desc
+
+
+def similarity_function(s1: str, s2: str) -> float:
+    # The ratio from s1,s2 may be different to s2,s1. We keep the max.
+    # See https://docs.python.org/3/library/difflib.html#difflib.SequenceMatcher.ratio
+    return max(SequenceMatcher(None, s1, s2).ratio(), SequenceMatcher(None, s2, s1).ratio())
+
+
+MIN_SIMIL_TO_CONSIDER = 0.7
+
+
+def suggest_similar(
+    requested_command: str, options: Iterable[str], similarity_function_to_use: Optional[Callable[[str, str], float]] = None
+) -> Optional[str]:
+    """
+    Given a requested command and an iterable of possible options returns the most similar (if any is similar)
+
+    :param requested_command: The command entered by the user
+    :param options: The list of available commands to search for the most similar
+    :param similarity_function_to_use: An optional callable to use to compare commands
+    :return: The most similar command or None if no one is similar
+    """
+
+    proposed_command = None
+    best_simil = MIN_SIMIL_TO_CONSIDER
+    requested_command_to_compare = requested_command.lower()
+    similarity_function_to_use = similarity_function_to_use or similarity_function
+    for each in options:
+        simil = similarity_function_to_use(each.lower(), requested_command_to_compare)
+        if best_simil < simil:
+            best_simil = simil
+            proposed_command = each
+    return proposed_command
