@@ -3,6 +3,7 @@
 """
 Cmd2 unit/functional testing
 """
+
 import builtins
 import io
 import os
@@ -28,6 +29,9 @@ from cmd2 import (
     plugin,
     utils,
 )
+from cmd2.rl_utils import (
+    readline,  # This ensures gnureadline is used in macOS tests
+)
 
 from .conftest import (
     HELP_HISTORY,
@@ -75,16 +79,20 @@ def test_version(base_app):
     assert cmd2.__version__
 
 
-@pytest.mark.skipif(sys.version_info >= (3, 8), reason="failing in CI systems for Python 3.8 and 3.9")
 def test_not_in_main_thread(base_app, capsys):
     import threading
 
-    cli_thread = threading.Thread(name='cli_thread', target=base_app.cmdloop)
+    # Mock threading.main_thread() to return our fake thread
+    saved_main_thread = threading.main_thread
+    fake_main = threading.Thread()
+    threading.main_thread = mock.MagicMock(name='main_thread', return_value=fake_main)
 
-    cli_thread.start()
-    cli_thread.join()
-    out, err = capsys.readouterr()
-    assert "cmdloop must be run in the main thread" in err
+    with pytest.raises(RuntimeError) as excinfo:
+        base_app.cmdloop()
+
+    # Restore threading.main_thread()
+    threading.main_thread = saved_main_thread
+    assert "cmdloop must be run in the main thread" in str(excinfo.value)
 
 
 def test_empty_statement(base_app):
@@ -141,7 +149,7 @@ def test_base_shortcuts(base_app):
 
 def test_command_starts_with_shortcut():
     with pytest.raises(ValueError) as excinfo:
-        app = cmd2.Cmd(shortcuts={'help': 'fake'})
+        cmd2.Cmd(shortcuts={'help': 'fake'})
     assert "Invalid command name 'help'" in str(excinfo.value)
 
 
@@ -171,9 +179,9 @@ now: True
     out, err = run_cmd(base_app, 'set quiet')
     expected = normalize(
         """
-Name   Value                           Description                                                 
+Name   Value                           Description
 ===================================================================================================
-quiet  True                            Don't print nonessential feedback                           
+quiet  True                            Don't print nonessential feedback
 """
     )
     assert out == expected
@@ -323,6 +331,16 @@ def test_base_error(base_app):
     assert "is not a recognized command" in err[0]
 
 
+def test_base_error_suggest_command(base_app):
+    try:
+        old_suggest_similar_command = base_app.suggest_similar_command
+        base_app.suggest_similar_command = True
+        out, err = run_cmd(base_app, 'historic')
+        assert "history" in err[1]
+    finally:
+        base_app.suggest_similar_command = old_suggest_similar_command
+
+
 def test_run_script(base_app, request):
     test_dir = os.path.dirname(request.module.__file__)
     filename = os.path.join(test_dir, 'script.txt')
@@ -427,6 +445,27 @@ def test_run_script_with_utf8_file(base_app, request):
     assert script_err == manual_err
 
 
+def test_scripts_add_to_history(base_app, request):
+    test_dir = os.path.dirname(request.module.__file__)
+    filename = os.path.join(test_dir, 'scripts', 'help.txt')
+    command = f'run_script {filename}'
+
+    # Add to history
+    base_app.scripts_add_to_history = True
+    base_app.history.clear()
+    run_cmd(base_app, command)
+    assert len(base_app.history) == 2
+    assert base_app.history.get(1).raw == command
+    assert base_app.history.get(2).raw == 'help -v'
+
+    # Do not add to history
+    base_app.scripts_add_to_history = False
+    base_app.history.clear()
+    run_cmd(base_app, command)
+    assert len(base_app.history) == 1
+    assert base_app.history.get(1).raw == command
+
+
 def test_run_script_nested_run_scripts(base_app, request):
     # Verify that running a script with nested run_script commands works correctly,
     # and runs the nested script commands in the correct order.
@@ -730,7 +769,21 @@ def test_pipe_to_shell_error(base_app):
     assert "Pipe process exited with code" in err[0]
 
 
-@pytest.mark.skipif(not clipboard.can_clip, reason="Pyperclip could not find a copy/paste mechanism for your system")
+try:
+    # try getting the contents of the clipboard
+    _ = clipboard.get_paste_buffer()
+    # pyperclip raises at least the following types of exceptions
+    #   FileNotFoundError on Windows Subsystem for Linux (WSL) when Windows paths are removed from $PATH
+    #   ValueError for headless Linux systems without Gtk installed
+    #   AssertionError can be raised by paste_klipper().
+    #   PyperclipException for pyperclip-specific exceptions
+except Exception:
+    can_paste = False
+else:
+    can_paste = True
+
+
+@pytest.mark.skipif(not can_paste, reason="Pyperclip could not find a copy/paste mechanism for your system")
 def test_send_to_paste_buffer(base_app):
     # Test writing to the PasteBuffer/Clipboard
     run_cmd(base_app, 'help >')
@@ -744,6 +797,38 @@ def test_send_to_paste_buffer(base_app):
     assert len(appended_contents) > len(paste_contents)
 
 
+def test_get_paste_buffer_exception(base_app, mocker, capsys):
+    # Force get_paste_buffer to throw an exception
+    pastemock = mocker.patch('pyperclip.paste')
+    pastemock.side_effect = ValueError('foo')
+
+    # Redirect command output to the clipboard
+    base_app.onecmd_plus_hooks('help > ')
+
+    # Make sure we got the exception output
+    out, err = capsys.readouterr()
+    assert out == ''
+    # this just checks that cmd2 is surfacing whatever error gets raised by pyperclip.paste
+    assert 'ValueError' in err and 'foo' in err
+
+
+def test_allow_clipboard_initializer(base_app):
+    assert base_app.allow_clipboard is True
+    noclipcmd = cmd2.Cmd(allow_clipboard=False)
+    assert noclipcmd.allow_clipboard is False
+
+
+# if clipboard access is not allowed, cmd2 should check that first
+# before it tries to do anything with pyperclip, that's why we can
+# safely run this test without skipping it if pyperclip doesn't
+# work in the test environment, like we do for test_send_to_paste_buffer()
+def test_allow_clipboard(base_app):
+    base_app.allow_clipboard = False
+    out, err = run_cmd(base_app, 'help >')
+    assert not out
+    assert "Clipboard access not allowed" in err
+
+
 def test_base_timing(base_app):
     base_app.feedback_to_output = False
     out, err = run_cmd(base_app, 'set timing True')
@@ -770,9 +855,7 @@ def _expected_no_editor_error():
         """
 EXCEPTION of type '{}' occurred with message: Please use 'set editor' to specify your text editing program of choice.
 To enable full traceback, run the following command: 'set debug true'
-""".format(
-            expected_exception
-        )
+""".format(expected_exception)
     )
 
     return expected_text
@@ -982,6 +1065,17 @@ def test_raise_keyboard_interrupt(base_app):
     assert 'Got a keyboard interrupt' in str(excinfo.value)
 
 
+@pytest.mark.skipif(sys.platform.startswith('win'), reason="SIGTERM only handled on Linux/Mac")
+def test_termination_signal_handler(base_app):
+    with pytest.raises(SystemExit) as excinfo:
+        base_app.termination_signal_handler(signal.SIGHUP, 1)
+    assert excinfo.value.code == signal.SIGHUP + 128
+
+    with pytest.raises(SystemExit) as excinfo:
+        base_app.termination_signal_handler(signal.SIGTERM, 1)
+    assert excinfo.value.code == signal.SIGTERM + 128
+
+
 class HookFailureApp(cmd2.Cmd):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -1264,9 +1358,7 @@ def test_select_options(select_app, monkeypatch):
    1. sweet
    2. salty
 {} with salty sauce, yum!
-""".format(
-            food
-        )
+""".format(food)
     )
 
     # Make sure our mock was called with the expected arguments
@@ -1292,9 +1384,7 @@ def test_select_invalid_option_too_big(select_app, monkeypatch):
    2. salty
 '3' isn't a valid choice. Pick a number between 1 and 2:
 {} with sweet sauce, yum!
-""".format(
-            food
-        )
+""".format(food)
     )
 
     # Make sure our mock was called exactly twice with the expected arguments
@@ -1323,9 +1413,7 @@ def test_select_invalid_option_too_small(select_app, monkeypatch):
    2. salty
 '0' isn't a valid choice. Pick a number between 1 and 2:
 {} with sweet sauce, yum!
-""".format(
-            food
-        )
+""".format(food)
     )
 
     # Make sure our mock was called exactly twice with the expected arguments
@@ -1349,9 +1437,7 @@ def test_select_list_of_strings(select_app, monkeypatch):
    1. math
    2. science
 Good luck learning {}!
-""".format(
-            'science'
-        )
+""".format('science')
     )
 
     # Make sure our mock was called with the expected arguments
@@ -1372,9 +1458,7 @@ def test_select_list_of_tuples(select_app, monkeypatch):
    1. Netflix
    2. WebSurfing
 Have fun procrasinating with {}!
-""".format(
-            'YouTube'
-        )
+""".format('YouTube')
     )
 
     # Make sure our mock was called with the expected arguments
@@ -1395,9 +1479,7 @@ def test_select_uneven_list_of_tuples(select_app, monkeypatch):
    1. Electric Guitar
    2. Drums
 Charm us with the {}...
-""".format(
-            'Drums'
-        )
+""".format('Drums')
     )
 
     # Make sure our mock was called with the expected arguments
@@ -1427,9 +1509,7 @@ def test_select_return_type(select_app, monkeypatch, selection, type_str):
    2. String
    3. Method
 The return type is {}
-""".format(
-            type_str
-        )
+""".format(type_str)
     )
 
     # Make sure our mock was called with the expected arguments
@@ -1566,17 +1646,98 @@ def test_multiline_input_line_to_statement(multiline_app):
     assert statement.multiline_command == 'orate'
 
 
-def test_clipboard_failure(base_app, capsys):
-    # Force cmd2 clipboard to be disabled
-    base_app._can_clip = False
+def test_multiline_history_no_prior_history(multiline_app):
+    # Test no existing history prior to typing the command
+    m = mock.MagicMock(name='input', side_effect=['person', '\n'])
+    builtins.input = m
 
-    # Redirect command output to the clipboard when a clipboard isn't present
-    base_app.onecmd_plus_hooks('help > ')
+    # Set orig_rl_history_length to 0 before the first line is typed.
+    readline.clear_history()
+    orig_rl_history_length = readline.get_current_history_length()
 
-    # Make sure we got the error output
-    out, err = capsys.readouterr()
-    assert out == ''
-    assert 'Cannot redirect to paste buffer;' in err and 'pyperclip' in err
+    line = "orate hi"
+    readline.add_history(line)
+    multiline_app._complete_statement(line, orig_rl_history_length=orig_rl_history_length)
+
+    assert readline.get_current_history_length() == orig_rl_history_length + 1
+    assert readline.get_history_item(1) == "orate hi person"
+
+
+def test_multiline_history_first_line_matches_prev_entry(multiline_app):
+    # Test when first line of multiline command matches previous history entry
+    m = mock.MagicMock(name='input', side_effect=['person', '\n'])
+    builtins.input = m
+
+    # Since the first line of our command matches the previous entry,
+    # orig_rl_history_length is set before the first line is typed.
+    line = "orate hi"
+    readline.clear_history()
+    readline.add_history(line)
+    orig_rl_history_length = readline.get_current_history_length()
+
+    multiline_app._complete_statement(line, orig_rl_history_length=orig_rl_history_length)
+
+    assert readline.get_current_history_length() == orig_rl_history_length + 1
+    assert readline.get_history_item(1) == line
+    assert readline.get_history_item(2) == "orate hi person"
+
+
+def test_multiline_history_matches_prev_entry(multiline_app):
+    # Test combined multiline command that matches previous history entry
+    m = mock.MagicMock(name='input', side_effect=['person', '\n'])
+    builtins.input = m
+
+    readline.clear_history()
+    readline.add_history("orate hi person")
+    orig_rl_history_length = readline.get_current_history_length()
+
+    line = "orate hi"
+    readline.add_history(line)
+    multiline_app._complete_statement(line, orig_rl_history_length=orig_rl_history_length)
+
+    # Since it matches the previous history item, nothing was added to readline history
+    assert readline.get_current_history_length() == orig_rl_history_length
+    assert readline.get_history_item(1) == "orate hi person"
+
+
+def test_multiline_history_does_not_match_prev_entry(multiline_app):
+    # Test combined multiline command that does not match previous history entry
+    m = mock.MagicMock(name='input', side_effect=['person', '\n'])
+    builtins.input = m
+
+    readline.clear_history()
+    readline.add_history("no match")
+    orig_rl_history_length = readline.get_current_history_length()
+
+    line = "orate hi"
+    readline.add_history(line)
+    multiline_app._complete_statement(line, orig_rl_history_length=orig_rl_history_length)
+
+    # Since it doesn't match the previous history item, it was added to readline history
+    assert readline.get_current_history_length() == orig_rl_history_length + 1
+    assert readline.get_history_item(1) == "no match"
+    assert readline.get_history_item(2) == "orate hi person"
+
+
+def test_multiline_history_with_quotes(multiline_app):
+    # Test combined multiline command with quotes
+    m = mock.MagicMock(name='input', side_effect=['  and spaces  ', ' "', ' in', 'quotes.', ';'])
+    builtins.input = m
+
+    readline.clear_history()
+    orig_rl_history_length = readline.get_current_history_length()
+
+    line = 'orate Look, "There are newlines'
+    readline.add_history(line)
+    multiline_app._complete_statement(line, orig_rl_history_length=orig_rl_history_length)
+
+    # Since spaces and newlines in quotes are preserved, this history entry spans multiple lines.
+    assert readline.get_current_history_length() == orig_rl_history_length + 1
+
+    history_lines = readline.get_history_item(1).splitlines()
+    assert history_lines[0] == 'orate Look, "There are newlines'
+    assert history_lines[1] == '  and spaces  '
+    assert history_lines[2] == ' " in quotes.;'
 
 
 class CommandResultApp(cmd2.Cmd):
@@ -1673,8 +1834,6 @@ def test_read_input_rawinput_true(capsys, monkeypatch):
         assert line == input_str
 
         # Run custom history code
-        import readline
-
         readline.add_history('old_history')
         custom_history = ['cmd1', 'cmd2']
         line = app.read_input(prompt_str, history=custom_history, completion_mode=cmd2.CompletionMode.NONE)
@@ -1821,7 +1980,7 @@ def test_poutput_none(outsim_app):
 def test_poutput_ansi_always(outsim_app):
     msg = 'Hello World'
     colored_msg = ansi.style(msg, fg=ansi.Fg.CYAN)
-    outsim_app.poutput(colored_msg)
+    outsim_app.poutput(colored_msg, apply_style=False)
     out = outsim_app.stdout.getvalue()
     expected = colored_msg + '\n'
     assert colored_msg != msg
@@ -1832,7 +1991,7 @@ def test_poutput_ansi_always(outsim_app):
 def test_poutput_ansi_never(outsim_app):
     msg = 'Hello World'
     colored_msg = ansi.style(msg, fg=ansi.Fg.CYAN)
-    outsim_app.poutput(colored_msg)
+    outsim_app.poutput(colored_msg, apply_style=False)
     out = outsim_app.stdout.getvalue()
     expected = msg + '\n'
     assert colored_msg != msg
@@ -2273,24 +2432,6 @@ def test_perror_no_style(base_app, capsys):
     assert err == msg + end
 
 
-@with_ansi_style(ansi.AllowStyle.ALWAYS)
-def test_pwarning_style(base_app, capsys):
-    msg = 'testing...'
-    end = '\n'
-    base_app.pwarning(msg)
-    out, err = capsys.readouterr()
-    assert err == ansi.style_warning(msg) + end
-
-
-@with_ansi_style(ansi.AllowStyle.ALWAYS)
-def test_pwarning_no_style(base_app, capsys):
-    msg = 'testing...'
-    end = '\n'
-    base_app.pwarning(msg, apply_style=False)
-    out, err = capsys.readouterr()
-    assert err == msg + end
-
-
 @with_ansi_style(ansi.AllowStyle.ALWAYS)
 def test_pexcept_style(base_app, capsys):
     msg = Exception('testing...')
@@ -2374,12 +2515,12 @@ def test_parseline_empty(base_app):
     assert not line
 
 
-def test_parseline(base_app):
+def test_parseline_quoted(base_app):
     statement = " command with 'partially completed quotes  "
     command, args, line = base_app.parseline(statement)
     assert command == 'command'
-    assert args == "with 'partially completed quotes"
-    assert line == statement.strip()
+    assert args == "with 'partially completed quotes  "
+    assert line == statement.lstrip()
 
 
 def test_onecmd_raw_str_continue(outsim_app):
